/*
 * Export Ben NanoNote display operations as a server.
 *
 * Copyright (C) 2018, 2020, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4740.h>
#include <l4/devices/lcd-jz4740-config.h>
#include <l4/devices/lcd-jz4740-panel.h>
#include <l4/devices/memory.h>

#include <l4/re/env.h>

#include <ipc/server.h>

#include "activation_client.h"
#include "activation_server.h"



/* Virtual address for the GPIO register block. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;

/* GPIO abstraction. */

static Gpio_jz4740_chip *gpio_port_c = 0;



/* GPIO pin definitions. */

enum Jz4740_lcd_gpio_mask
{
  Jz4740_lcd_gpio_d0_d7 = 0xff,
};

enum Jz4740_lcd_gpio
{
  Jz4740_lcd_gpio_clk = 18, /* pixel clock */
  Jz4740_lcd_gpio_rs  = 19, /* hsync; command/data select */
  Jz4740_lcd_gpio_cs  = 20, /* vsync; chip select */
};



static int setup_memory(void)
{
  if (get_memory("jz4740-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  return 0;
}



/* Display device only implementing the activation interface. */

class DisplayObject_server : public Activation
{
  Pin_slice slcd8_mask = {.offset=0, .mask=(1 << Jz4740_lcd_gpio_cs) | (1 << Jz4740_lcd_gpio_rs) |
                                           (1 << Jz4740_lcd_gpio_clk) | Jz4740_lcd_gpio_d0_d7};

  Activation *_backlight;

public:
  explicit DisplayObject_server(Activation *backlight)
  : _backlight(backlight)
  {
  }

  /* Switch the display off. */

  long disable(void)
  {
    /* Configure SLCD8 pins. */

    gpio_port_c->multi_setup(slcd8_mask, Hw::Gpio_chip::Input, 0);
    _backlight->disable();
    return L4_EOK;
  }

  /* Switch the display on. */

  long enable(void)
  {
    /* Configure SLCD8 pins. */

    gpio_port_c->multi_config_pad(slcd8_mask, Hw::Gpio_chip::Function_alt, 0);
    _backlight->enable();
    return L4_EOK;
  }
};



int main(void)
{
  if (setup_memory())
    return 1;

  /* Initialise the GPIO abstraction. */

  Gpio_jz4740_chip gpio_port(gpio_virt_base + 0x200, gpio_virt_base + 0x300, 32);

  gpio_port_c = &gpio_port;

  /* Obtain a reference to the backlight device. */

  l4_cap_idx_t backlight = l4re_env_get_cap("backlight");

  if (!l4_is_valid_cap(backlight))
    return 1;

  /* Use the activation interface with the backlight. */

  client_Activation backlight_obj(backlight);

  /* Initialise and register a new server object. */

  DisplayObject_server obj(&backlight_obj);

  /* Bind and start the IPC server loop. */

  if (ipc_server_loop_for(Activation, &obj, "display"))
    return 1;

  return 0;
}
